#!/usr/bin/perl -w

use strict "vars";
use Getopt::Std;

my $usage;
{
$usage = <<"_USAGE_";
This script has been written for the property generation task of the
ESSLLI 2008 Lexical Semantics workshop, where the gold standard is
based on the feature norms of McRae et al. 2005.

Given a gold standard of concept-property pairs and the output of a
word space model, the script computes (cross-concept) average
precision for n-best lists of various sizes. The gold standard is
expected to contain one or more "expansions" of each property, i.e.,
potential realizations of the property in a format that is more
suitable for word space models, e.g., a property such as
"makes_loud_noise" might be "expanded" into "noise", "loud", "noisy",
etc.

The model output is matched against the expansions, so that a
(case-insensitive) match with any of the expansions counts as a hit
for the relevant property.

The script takes as input a (space or tab-delimited) "expansion" file
in format:

concept true_property expansion1 (... expansionN)

and a (space or tab-delimited) "model output" in format:

concept property score

The user can also pass a comma-delimited list of n-best "thresholds"
with the option -t. These thresholds determine the sizes of the n-best
lists -- e.g., if the user passes "-t 10,20" the script will return
the average of the precisions computed for the 10-best and 20-best
lists associated to each concept. If no list is passed, default
thresholds are 10-, 20- and 30-best.

The script computes precision for each target concept in the expansion
file (concepts not in the expansion file but present in the model
output are ignored) at each of the specified (or default) n-best
thresholds, and then it reports the average precision across concepts
for each threshold. If there are ties in the model scores, they are
broken randomly. Moreover, if multiple expansions of a true property
are matched by model-generated properties, only the top-ranked match
is considered, and the lower ones are ignored (i.e., they are not
treated as hits, but they do not contribute to the n-count either). On
the other hand, the script assumes that the model-generated properties
of a single concept are unique -- repeated properties will lead the
script to assign the lowest score to all repeated instances and
probably to other unwanted behaviour.

The script only computes precision since this is the variable of most
interest in the ESSLLI task, where concepts have (almost) the same
number of golden standard properties, and recall can be easily
inferred from precision.

For each threshold level, the model outputs the following information:

threshold concept_number precision_average precision_sd

The concept_number field records the number of concepts that had
enough properties to compute precision at the current threshold (e.g.,
for 10-best lists, the number of concepts that had at least 10
model-generated properties). Please pay great attention to this
number, especially when comparing the output of different models, a
comparison that will be meaningless if the average n-best precision of
a model is based on a lot less concept lists than the one of another.

The following parameters/options can be specified:

-h: print this documentation and exit

-t: comma delimited list of threshold values at which to compute a
 n-best list (default: 10,20,30)

Usage examples:

evaluate-against-expanded-props.pl -h | more

evaluate-against-expanded-props.pl -t 10,100 explist.txt model.txt

evaluate-against-expanded-props.pl explist.txt model.txt

Copyright 2007, Marco Baroni

This program is free software. You may copy or redistribute it under
the same terms as Perl itself.

Please let me know if you find problems/errors/bugs in this script!
_USAGE_
}
{
    my $blah = 1;
# this useless block is here because here document confuses
# emacs
}

##################### initialization #####################

my %opts = ();

getopts('ht:',\%opts);

if ($opts{h}) {
    print $usage;
    exit;
}


# statistics will be computed by default for top 10, 20 and 30
# properties, but user can pass comma-delimited list
my @thresholds =  (10,20,30);
if ($opts{t}) {
    @thresholds =  split ",",$opts{t};
}
# sort, just in case
@thresholds = sort {$a <=> $b} @thresholds;
# from user-friendly values to array indices
my $i = 0;
while (defined($thresholds[$i])) {
    $thresholds[$i]--;
    $i++;
}

# the input files
my $expansion_file = shift;
my $score_file = shift;



# various data structures we will need later
my %exp_to_true_property = ();
my @concepts = ();
my %seen_concepts = ();
my %raw_properties = ();
my %scores = ();
my %random_values = ();
my %seen_random = ();
my %precisions = ();
my %squared_precisions = ();
my %enough_properties = ();

##################### end of initialization #####################


##################### reading expansion file #####################

open EXPANSIONS,$expansion_file  or
    die "could not find expansion file $expansion_file";
while (<EXPANSIONS>) {
    chomp;

    # field 0 contains concept, field 1 contains mcrae prop,
    # everything else is expansions
    my @F = split "[\t ]+",$_;

    # for pairs concept-expansion (lower-cased), we store corresponding mcrae
    # prop in double hash %exp_to_true_property
    foreach my $expansion (@F[2..$#F]) {
	$exp_to_true_property{$F[0]}{lc($expansion)} = $F[1];
    }
}
close EXPANSIONS;

##################### end of expansion file reading #####################


##################### reading algorithm scores #####################

open SCORES,$score_file  or
    die "could not find score file $score_file";
while (<SCORES>) {
    chomp;
    my ($concept,$property,$score) = split "[\t ]+",$_;
    
    # we store any unseen concept in the concepts array
    if (!$seen_concepts{$concept}++) {
	push @concepts,$concept;
    }

    # lower-casing the property
    $property = lc($property);

    # we push the property in an array of properties for the current
    # concept
    push @{$raw_properties{$concept}},$property;

    # keep track of score for current concept-property pair
    $scores{$concept}{$property} = $score;

    # we also generate a unique random number per concept-property
    # pair, to break ties later
    my $random_value = rand();
    while ($seen_random{$random_value}++) {
	$random_value = rand();
    }
    $random_values{$concept}{$property} = $random_value;

}
close SCORES;

##################### end of algorithm score reading #####################


##################### compute concept-specific precisions #####################

# we process one concept at the time
foreach my $concept (@concepts) {
 
    # we sort the properties by the scores, using the
    # unique random values to break ties
    my @sorted_props = sort 
    {
	$scores{$concept}{$b} <=> $scores{$concept}{$a} ||
	    $random_values{$concept}{$b} <=> $random_values{$concept}{$a}
    }
    @{$raw_properties{$concept}};

    # an index to traverse the threshold array
    my $threshold_index = 0;
    # an index to count the properties we keep
    my $property_index = 0;

    # we will ignore duplicate "hits" due to more than one model
    # property matching expansions of the same true property
    my %seen_properties = ();

    my $hits = 0;

    # traversing the property array
    foreach my $property (@sorted_props) {

	# 3 possibilities:
	# expansion of unseen property: count as hit;
	# expansion of seen property: ignore;
	# no expansion: count as miss;

	if (my $true_property = $exp_to_true_property{$concept}{$property}) {
	    # ignore previously seen properties
	    if ($seen_properties{$true_property}++) {
		next;
	    }
	    # else, we saw a true property, so this is a hit
	    $hits++;
	}

	# are we at a threshold?
	if ($property_index == $thresholds[$threshold_index]) {
	    # if so, we use the number of hits at this point to
	    # compute precision at current threshold, and add it
	    # to the overall count for threshold
	    my $precision = $hits/($property_index+1);
	    $precisions{$property_index} += $precision;
	    # we also need squared precisions to compute variance
	    $squared_precisions{$property_index} += $precision**2;

	    # ... and we keep track of the fact that this is another
	    # concept with the relevant number of properties
	    $enough_properties{$property_index}++;
	    
	    # we increase the threshold index...
	    $threshold_index++;
	    # ... and if we traversed the whole thresholds array, we
	    # are actually done with the current concept
	    if ($threshold_index > $#thresholds) {
		last;
	    }

	}
	# if we are not done with concept, we look at the next
	# model-generated property
	$property_index++;
    }
    
}    

##################### concept-specific precisions computed #####################

##################### averaging precisions #####################

# OK, finally we compute average of precisions, and their s.d.'s
foreach my $threshold (@thresholds) {

    print $threshold+1,"\t";

    my $prec_avg = 0;
    my $prec_sd = 0;

    # we must account for the (hopefully highly unlikely) case in
    # which there was no concept with at least n properties
    my $n = 0;
    if ($n = $enough_properties{$threshold}) {
	$prec_avg = $precisions{$threshold}/$n;
	# we use the "computational" form of variance:
	# E(X^2) - (E(X))^2
	# (note that since we are not interested in statistical
	# inference we use n rather than n-1)
	my $prec_var = ($squared_precisions{$threshold}/$n) - ($prec_avg**2);
	$prec_sd = sqrt($prec_var);
    }

    print $n,"\t",$prec_avg,"\t",$prec_sd,"\n";
}

##################### averaging precisions #####################

